/*
   bugger 0.01b - Linux process bugger
   -----------------------------------
   Copyright (C) 2003 by Michal Zalewski <lcamtuf@coredump.cx>

   Licensed under terms and conditions of GNU Public License version 2.

 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ptrace.h>
#include <fcntl.h>
#include <string.h>
#include <sys/wait.h>
#include <signal.h>
#include <time.h>
#include <sys/mman.h>
#include <asm/page.h>
#include <getopt.h>
#include <errno.h>
#include "types.h"

#define CODESEG		0x08
#define DEBUG_TRAP	0x08048000
#define MAXVAL 		100000000

// Macros: r_every(n): 1 every n insns on average
//         r_avg(n): a random between 0 and 2n
//         r_rnd(n) : a random between 0 and n-1
//         r_percent(n): 1 every n out of 100 insns on average


#define r_rnd(n)	(rand() % (n))
#define r_avg(n)	(r_rnd(((n) << 1) + 1))
#define r_every(n)      (r_rnd(n) == 0)
#define r_percent(n)    (r_rnd(100) < (n))


#define detach() do { \
    if (cpid > 0) { \
      ptrace(PTRACE_DETACH,cpid,0,0); \
      usleep(100000); \
      if (!kill(cpid,SIGCONT)) debug("Child process set free...\n"); \
    } \
  } while (0)

#define debug(x...) fprintf(stderr,x)

#define fatal(x...) do { \
     fprintf(stderr,"[-] ERROR: " x); \
     kill(cpid,SIGKILL); \
     exit(1); \
   } while (0)


static _u32 
     mod_skip = 0,
     mod_dist = 3000,
     mod_insns = 1000,  
     max_insert = 4096,

     prob_flags = 30,
     prob_reg = 20,

     prob_stack = 10,
     prob_data = 70,
     prob_heap = 30,
     prob_int = 90,

     prob_follow = 50,
     prob_insert = 5;

_u32 cpid;


static void sighand(_s32 x) {
  debug("+++ Caught signal %d - bailing out +++\n",x);
  fatal("user signal received.\n");
  exit(1);
}


static void prepare_process(char** argv) {

  _s32 w;

  if (!(cpid=fork())) {
    ptrace(PTRACE_TRACEME,0,0,0);
    execvp(argv[0],&argv[0]);
    fatal("execvp(%s) failed.\n",argv[0]);
  }

  if (cpid < 0) fatal("fork() failed.\n");

  if (waitpid(cpid,&w,WUNTRACED) < 0) fatal("waitpid() failed.\n");

  if (WIFEXITED(w)) exit(1);
  if (WIFSIGNALED(w)) fatal("child died on signal %d.\n",WTERMSIG(w));
  if (WSTOPSIG(w) != SIGTRAP) fatal("unknown child stop condition.\n");

  debug("Child attached successfully at execve() - PID %d.\n",cpid);

}


static void usage(_u8* myname) {
  debug("Usage: %s [ -S n ] [ -D n ] [ -C n ] [ -x n ] [ -f n ] [ -r n ]\n"
        "  [ -t n ] [ -d n ] [ -m n ] [ -i n ] [ -P n ] [ -s n ] [ -- ] \n"
        "  command [ params... ]\n\n"
        "Modification interval / parameter control:\n"
        "  -S n  - skip n first instructions (%u)\n"
        "  -D n  - modification at distance n (%u)\n"
        "  -C n  - modification after n instructions (%u)\n"
        "  -x n  - max. string insert length (%u bytes)\n\n"
        "Object modification probabilities (percents, 0 to disable):\n"
        "  -f n  - comparison result modification probability (%u%%)\n"
        "  -r n  - register value modification probability (%u%%)\n"
        "  -t n  - stack modification probability (%u%%)\n"
        "  -d n  - data modification probability (%u%%)\n"
        "  -m n  - heap modification probability (%u%%)\n"
        "  -i n  - integer modification probability (%u%%)\n"
        "  -P n  - pointer follow probability (%u%%)\n"
        "  -s n  - string insertion probability (%u%%)\n\n",
        myname,mod_skip,mod_dist,mod_insns,max_insert,prob_flags,prob_reg,
	prob_stack,prob_data,prob_heap,prob_int,prob_follow,prob_insert);

  debug("IF YOU ARE NOT SURE WHAT THIS PROGRAM DOES, STOP NOW AND READ THE\n"
        "DOCUMENTATION. FAILURE TO DO SO MAY RESULT IN DATA LOSS!\n\n");

  exit(1);
}


#define CHECKPROC(r) do {\
    if (waitpid(cpid,&w,WUNTRACED) < 0 || \
        (WIFSTOPPED(w) && WSTOPSIG(w) != SIGTRAP) || \
        WIFSIGNALED(w) || WIFEXITED(w)) \
      fatal("child process gone or signaled " r ".\n"); \
  } while (0)


_u32 data_start,data_len;

static void bug_start(void) {
  _u8 buf[1024];
  _s32 f,w;
  _u32 i,v;
  FILE* F;

  sprintf(buf,"/proc/%d/maps",cpid);
  F=fopen(buf,"r");
  if (!F) fatal("cannot open %s.\n",buf);
  if (!fgets(buf,sizeof(buf),F) || !fgets(buf,sizeof(buf),F))
    fatal("cannot read data from /proc/pid/maps.\n");
  fclose(F);

  if (sscanf(buf,"%x-%x",&data_start,&data_len)!=2)
    fatal("bad /proc/pid/maps format.\n");

  data_len -= data_start;

  debug("[+] Data start 0x%x length %d bytes.\n",data_start,data_len);

  sprintf(buf,"/proc/%d/exe",cpid);
  f=open(buf,O_RDONLY);
  if (f<0) fatal("cannot open %s.\n",buf);
  lseek(f,24,SEEK_SET);
  read(f,&i,4);
  close(f);

  if (i >> 24 != CODESEG) fatal("code address (0x%x) unsupported.\n",i);
  v=ptrace(PTRACE_PEEKDATA,cpid,i,0);
  if (v != 0x895eed31) fatal("unrecognized _start prolog.\n");

  debug("[+] Found known _start (0x%x), skipping libc prolog...\n",i);

  ptrace(PTRACE_POKEDATA,cpid,DEBUG_TRAP,0xe9cc); /* int3, jmpl .... */
  ptrace(PTRACE_POKEDATA,cpid,i,0x895e90cc);
  ptrace(PTRACE_CONT,cpid,0,0);

  CHECKPROC("before _start");

}


static void go_main(void) {

  _u32 cyc=0,eip;
  _s32 w;

  debug("[+] Now at _start, looking for main()...\n");

  while (cyc++ < 100) {
    _u8 cmd;

    ptrace(PTRACE_SINGLESTEP,cpid,0,0);

    CHECKPROC("in _start");

    eip=ptrace(PTRACE_PEEKUSER,cpid,12*4 /* eip */,0);
    cmd=ptrace(PTRACE_PEEKDATA,cpid,eip,0);

    if (cmd == 0xe8) {     
      _u32 esp,val;
      esp=ptrace(PTRACE_PEEKUSER,cpid,15*4 /* esp */,0);
      val=ptrace(PTRACE_PEEKDATA,cpid,esp,0);
      debug("[+] Found main() at 0x%x, skipping _start...\n\n",val);
      ptrace(PTRACE_POKEDATA,cpid,DEBUG_TRAP+2,val - DEBUG_TRAP - 6);
      ptrace(PTRACE_POKEDATA,cpid,esp,DEBUG_TRAP);
      ptrace(PTRACE_CONT,cpid,0,0);
      return;
    }
  }

  fatal("no known main() reference from _start.\n");

}


#define CP(x)	( ((x) < 0 || (x) > 100) ? ({ \
                  fatal("invalid percentage.\n"); 0; }) : (x))

#define CV(x)	( ((x) < 0 || (x) > MAXVAL) ? ({ \
                  fatal("invalid parameter value.\n"); 0; }) : (x))

static void parse_options(_s32 argc,_u8** argv) {
  _s32 opt,seed;
 
  while ((opt=getopt(argc,(void*)argv,"+S:D:C:x:f:t:r:d:i:P:s:m:"))!=EOF) 
    switch (opt) {
      case 'S': mod_skip    = CV(atoi(optarg)); break;
      case 'D': mod_dist    = CV(atoi(optarg)); break;
      case 'C': mod_insns   = CV(atoi(optarg)); break;
      case 'x': max_insert  = CV(atoi(optarg)); break;
      case 'f': prob_flags  = CP(atoi(optarg)); break;
      case 'r': prob_reg    = CP(atoi(optarg)); break;
      case 't': prob_stack  = CP(atoi(optarg)); break;
      case 'd': prob_data   = CP(atoi(optarg)); break;
      case 'm': prob_heap   = CP(atoi(optarg)); break;
      case 'i': prob_int    = CP(atoi(optarg)); break;
      case 'P': prob_follow = CP(atoi(optarg)); break;
      case 's': prob_insert = CP(atoi(optarg)); break;
      default: usage(argv[0]);
    }

  if (!(argc-optind)) usage(argv[0]);
  
  seed = time(0) ^ (getpid() << 16);
  srand(seed);

  debug("\nCommand: %s\n",argv[optind]);
  debug("  skip:%u dist:%u insns:%u insert:%u seed:%x\n",mod_skip,mod_dist,
        mod_insns,max_insert,seed);

  debug("  fl:%u rx:%u st:%u dt:%u hp:%u int:%u pt:%u ins:%u (%%)\n\n",
        prob_flags,prob_reg,prob_stack,prob_data,prob_heap,
	prob_int,prob_follow,prob_insert);

}


static void modify(_u32 ad) {
  _u32 cur;
  cur  = ptrace(PTRACE_PEEKDATA,cpid,ad,0);
  // HANDLE INSERT! FIXME BLINK BLINK ** FIXME ** FIX!
  cur += r_rnd(max_insert) - (max_insert >> 1);
  ptrace(PTRACE_POKEDATA,cpid,ad,cur);
}
 

static inline _u8 is_pointer(_u32 x) {
  x >>= 24;
  if ((x >= 0x40 && x <= 0x45) || x == 0x08 || x == 0xbf) return 1;
  return 0;
}


int main(int argc,char** argv) {

  _u32 csig=0, cyc = 0, peip, eip = 0, distmax = 0, lmod = 0;
  _s32 w;

  debug("bugger 0.01b by Michal Zalewski <lcamtuf@coredump.cx>\n");

  parse_options(argc,(_u8**)argv);
  prepare_process(&argv[optind]);

  // Squeeak!
  signal(SIGINT,sighand);  signal(SIGIO,sighand);  signal(SIGHUP,sighand);
  signal(SIGQUIT,sighand); signal(SIGILL,sighand); signal(SIGBUS,sighand);
  signal(SIGABRT,sighand); signal(SIGFPE,sighand); signal(SIGPIPE,sighand);
  signal(SIGSEGV,sighand); signal(SIGTERM,sighand);

  bug_start();
  go_main();

  while (1) {

    if (waitpid(cpid,&w,WUNTRACED) < 0)
      fatal("child process killed to death!\n");

    if (WSTOPSIG(w) != SIGTRAP) {
      if (WIFSTOPPED(w))
        debug("--- Caught signal %d ---\n",csig=WSTOPSIG(w));
      else if (WIFEXITED(w)) {
        debug("+++ Exited with code %d (%u instructions) +++\n", 
              WEXITSTATUS(w),cyc);
        break;
      } else if (WIFSIGNALED(w)) {
        debug("+++ Exited on signal %d (%u instructions) +++\n",WTERMSIG(w),
              cyc);
        break;
      }
    }

    peip=eip;
    eip=ptrace(PTRACE_PEEKUSER,cpid,12*4 /* eip */,0);

    if (eip >> 24 != CODESEG) {      
      if (peip >> 24 == CODESEG) {

        _u32 esp,val;

        esp=ptrace(PTRACE_PEEKUSER,cpid,15*4 /* esp */,0);
        val=ptrace(PTRACE_PEEKDATA,cpid,esp,0);

        // Shift for PLT trampolines
        if (val >> 24 == 0x40) {
          esp+=8;
          val=ptrace(PTRACE_PEEKDATA,cpid,esp,0);
        }

        if (val >> 24 == CODESEG) {
          // Install backdoor
          ptrace(PTRACE_POKEDATA,cpid,DEBUG_TRAP+2,val - DEBUG_TRAP - 6);
          ptrace(PTRACE_POKEDATA,cpid,esp,DEBUG_TRAP);
          ptrace(PTRACE_CONT,cpid,0,0);
          continue;
        }
      }
    } else {

      _u32 ad = 0 /* gcc sucks */,rno,prev;
    
      cyc++; 
      
      if ( (cyc > mod_skip) && ( r_every(mod_insns) 
           || (abs(eip-lmod) > distmax))) {
	  
        lmod=eip;
        distmax = r_avg(mod_dist);
      
        if (r_percent(prob_flags)) {
          _u32 eflags;
  	  eflags=ptrace(PTRACE_PEEKUSER,cpid,14*4 /* eflags */,0);
          // Change (xor) only some flags:
	  // Bit:    CBA9876543210 (eflags)
	  // Flag:    OD  sz a p c - change only those
	  // Mask:   0110011010101 = 0xCD5
	  eflags ^= ( rand() & 0xCD5);
          ptrace(PTRACE_POKEUSER,cpid,14*4 /* eflags */,eflags);
        }
      
        if (r_percent(prob_reg)) {
          // Modify eax (6), ebx (0), ecx (1), edx (2), esi (3) or edi (4).
          _u32 val;
	  rno = r_rnd(6);
  	  if (rno == 5) rno=6; // ekip ebp, go to eax
          val=ptrace(PTRACE_PEEKUSER,cpid,4*rno,0);
          val += r_rnd(max_insert) - (max_insert >> 1);
	  ptrace(PTRACE_POKEUSER,cpid,4*rno,val);
        }
      
reselect:
      
        // Select a random data address, something near esp, or
        // a random register, granted its value looks like a pointer
        switch (r_rnd(3)) {
          case 0: ad = data_start + r_rnd(data_len); break;
          case 1: ad = ptrace(PTRACE_PEEKUSER,cpid,15*4 /* esp */,0);
                  ad += r_rnd(32) << 2; break;
	  case 2: rno = r_rnd(6);
                  if (rno == 5) rno=6;
                  ad = ptrace(PTRACE_PEEKUSER,cpid,4*rno,0);
                  if (!is_pointer(ad)) goto reselect;
                  ad += r_rnd(max_insert) - (max_insert >> 1);
                  break;
        }

        prev=ad;

        while (r_every(prob_follow) && is_pointer(ad)) {
          prev=ad;
          ad=ptrace(PTRACE_PEEKDATA,cpid,ad,0);
        }

        switch (ad >> 24) {
          case 0x40 ... 0x45: if (r_percent(prob_heap)) modify(ad); break;
          case 0x08: if (r_percent(prob_data)) modify(ad); break;
          case 0xbf: if (r_percent(prob_stack)) modify(ad); break;
          default: if (r_percent(prob_int)) modify(prev); break;
        }
	
      }
      
    }

    ptrace(PTRACE_SINGLESTEP,cpid,1,csig);
    csig=0;

  }

  detach();

  return 0;

}
